/*
 * Copyright (c) 2003 Communication Networks Lab, Simon Fraser University.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Authors: Steve Wei Shen <wshen@cs.sfu.ca>,  Z. Morley Mao.
 *
 */

#include "dampinfo.h"
#include "Timing/damp_reuse_timer.h"
#include "routeinfo.h"
#include "rtProtoBGP.h"
#include <math.h>

int DampInfo::flapNum = 0;
int DampInfo::supNum = 0;

/** Constructs damping information about a route */
DampInfo::DampInfo(RouteInfo* ri, rtProtoBGP* b, double wp, double cp, double sl, double rl, double hl, double mst) {
  //assume called during announcement
  bgp = b;
  routeInfo = ri;
  penalty = 0;
  lastUpdatedTime = bgp->now();
  suppressed = false;
  prevUpdate = 1;
  reuseTimer = NULL;
  prevDop = routeInfo->dop;
  twoBits = 0;
  peerNh = routeInfo->peer->nh;
  nlri = routeInfo->route->nlri;
  flapCount = 0;
  WDcount = 0;
  flapWDcount = 0;
  withdrawPenalty = wp;
  changePenalty = cp;
  suppressLimit = sl;
  reuseLimit = rl;
  halfLife_ok = hl;
  halfLife_ng = hl;
  maxSuppressTime = mst;
  penalty_ceiling = reuseLimit * exp(maxSuppressTime/halfLife_ok) * log(2.0);
}


/** Updates the penalty value of a route */
void DampInfo::updatePenalty(bool withdraw) {
  double now = bgp->now();
  bool reachable = (withdraw) ? false : true;
  //decay penalty before incrementing it
  double coeff;
  if(reachable) {
    coeff = log(2.0) / halfLife_ok;
  } else {
    coeff = log(2.0) / halfLife_ng;
  }
  double diff = now - lastUpdatedTime;
  penalty *= exp(-coeff*diff);
  lastUpdatedTime = now;
  if(bgp->selectiveRFD && WDcount != 0) {
    WDcount *= exp(-coeff*diff);
  }
}


/** Updates damping information  */
void DampInfo::update(bool withdraw, RouteInfo* ri) {
  //When to timeout the data structure, delete DampInfo from the dampedRoutes?
  //Set up a timer, when the penalty value reaches to some lower bound, remove it.
  bool reachable = (withdraw) ? false : true;
  //decay penalty before incrementing it
  updatePenalty(withdraw);
  //beginning of two added algorithms 
  if(bgp->rfdCom) {
    if(flapCount <= 2) {   //using rfdPP
      //printf("flapCount = %d\n", flapCount);
      if(withdraw && prevUpdate == 1) {
        flapWDcount = 1;
      }
      else if(!withdraw) {
        int curTwoBits;
        int curDop = ri->dop;
        if(curDop == prevDop) {
	  curTwoBits = 1;
        }
        else if(curDop > prevDop) {
	  curTwoBits = 2;
        }
        else {
          curTwoBits = 3;
        }
        if(prevUpdate == 0 && bgp->collectedRoutes[ri->peer][ri->route->nlri->val()]->find(ri->route) > 1) {
          penalty = penalty + withdrawPenalty;
	  flapCount = flapCount + 1;
	  bgp->collectedRoutes[ri->peer][ri->route->nlri->val()]->clear();
	  flapNum = flapNum + 1;
        }
        else if(prevUpdate == 1 && curTwoBits == 2 && bgp->collectedRoutes[ri->peer][ri->route->nlri->val()]->find(ri->route) > 1) {
          penalty = penalty + changePenalty;
	  flapCount = flapCount + 1;
          bgp->collectedRoutes[ri->peer][ri->route->nlri->val()]->clear();
	  flapNum = flapNum + 1;
        }
        else if(curTwoBits == 1 && flapWDcount == 1) {     // 'up down up' state is detected
	  penalty = penalty + withdrawPenalty;
	  flapCount = flapCount + 1;
	  flapNum = flapNum + 1;
          bgp->collectedRoutes[ri->peer][ri->route->nlri->val()]->clear();
        }  
        prevDop = curDop;
        flapWDcount = 0;
      }
    }
    else {    //using traditional RFD
      //printf("enter the traditional RFD algorithm\n");
      //flap:10
      if(withdraw && prevUpdate == 1) {
        penalty += withdrawPenalty;
        flapCount++;
        flapNum++;
      }
      //flap:11(implicit_withdrawal)
      else if(!withdraw && prevUpdate == 1) {
        penalty += changePenalty;
        flapCount++;
        flapNum++;
      }
      //flap:01
      else if(!withdraw && prevUpdate == 0) {
        //Cisco doesn't punish readvertisement after withdrawal
        if(bgp->rfd_punish_readvertisement) {  //if readvertisement after withdrawal is punished
	  penalty += withdrawPenalty; //the penalty is the same as a withdrawal
	  flapCount++;
          flapNum++;
        }
      }
    //  if(bgp->collectedRoutes[ri->peer][ri->route->nlri->val()] != NULL) {
    //    bgp->collectedRoutes[ri->peer][ri->route->nlri->val()]->clear();
    //	} 
    }
  }
  else if(bgp->rfdPP) {
    if(withdraw && prevUpdate == 1) {
      flapWDcount = 1;
    }
    else if(!withdraw) {
      int curTwoBits;
      int curDop = ri->dop;
      if(curDop == prevDop) {
	curTwoBits = 1;
      }
      else if(curDop > prevDop) {
	curTwoBits = 2;
      }
      else {
	curTwoBits = 3;
      }
      if(prevUpdate == 0 && bgp->collectedRoutes[ri->peer][ri->route->nlri->val()]->find(ri->route) > 1) {
        penalty = penalty + withdrawPenalty;
	flapCount = flapCount + 1;
	bgp->collectedRoutes[ri->peer][ri->route->nlri->val()]->clear();
	flapNum = flapNum + 1;
      }
      else if(prevUpdate == 1 && curTwoBits == 2 && bgp->collectedRoutes[ri->peer][ri->route->nlri->val()]->find(ri->route) > 1) {
        penalty = penalty + changePenalty;
	flapCount = flapCount + 1;
        bgp->collectedRoutes[ri->peer][ri->route->nlri->val()]->clear();
	flapNum = flapNum + 1;
      }
      else if(curTwoBits == 1 && flapWDcount == 1) {     // 'up down up' state is detected
	penalty = penalty + withdrawPenalty;
	flapCount = flapCount + 1;
	flapNum = flapNum + 1;
        bgp->collectedRoutes[ri->peer][ri->route->nlri->val()]->clear();
      }
      prevDop = curDop;
      flapWDcount = 0;
    }
  }   
  //end of two added algorithms
  else if(bgp->rfdPlus) {
    if(withdraw) {      //if it's a withdrawal
      prevDop = -1;     //set the dop of a withdrawal to be the smallest --- -1
    }
    else {   //if it's an advertisement
      int curDop = ri->dop;
      //in case the current route preference is higher than the previous one and the current route is in
      //the collectedRoutes
      if(curDop > prevDop && bgp->collectedRoutes[ri->peer][ri->route->nlri->val()]->find(ri->route) > 1) {
        if(prevUpdate == 1)  { //if previous update is also an advertisement
	  penalty = penalty + changePenalty;
	  flapCount = flapCount + 1;
	  bgp->collectedRoutes[ri->peer][ri->route->nlri->val()]->clear();
          //bgp->collectedRoutes[ri->peer][ri->route->nlri->val()]->add(ri->route);
	  flapNum = flapNum + 1;
	}  
        else {   //if previous update is a withdrawal
	  penalty = penalty + withdrawPenalty;
          flapCount = flapCount + 1;
          flapNum = flapNum + 1;
	  bgp->collectedRoutes[ri->peer][ri->route->nlri->val()]->clear();
          //bgp->collectedRoutes[ri->peer][ri->route->nlri->val()]->add(ri->route);
        }
      }
      prevDop = curDop;
    }
  }
  else if(bgp->selectiveRFD) {
    //flap:10, postpone considering this as a flap
    if(withdraw && prevUpdate == 1) {
      WDcount += withdrawPenalty;
      flapWDcount++;
    }
    else if(!withdraw) {
      int curTwoBits;
      int curDop = ri->dop;
      if(curDop == prevDop) {
	curTwoBits = 1;
      }
      else if(curDop > prevDop) {
	curTwoBits = 2;
      }
      else {
	curTwoBits = 3;
      }
      /* is a flap:
	 (1) route preference change has altered direction: 10->11 or 11->10
	 (2) up down up: 01 and WDcount != 0
      */
      /*
      if((twoBits > 1 && curTwoBits > 1 && twoBits != curTwoBits) ||
	 (curTwoBits == 1 && flapWDcount != 0)) {
	penalty = penalty + WDcount + changePenalty;
	flapCount = flapCount + flapWDcount + 1;
	} */
      //the above original code is modified (the following 13 lines of code were added)
      //in case route preference change has altered direction
      if(twoBits > 1 && curTwoBits > 1 && twoBits != curTwoBits) {
	penalty = penalty + WDcount + changePenalty;
	flapCount = flapCount + flapWDcount + 1;
	flapNum = flapNum + flapWDcount + 1;
      }
      //in case the 'up down up' situation occurs
      if(curTwoBits == 1 && flapWDcount != 0) {
        if(bgp->rfd_punish_readvertisement) {  //if readvertisement after withdrawal is punished
	  penalty = penalty + WDcount + withdrawPenalty;
	  flapCount = flapCount + flapWDcount + 1;
	  flapNum = flapNum + flapWDcount + 1;
	} else {  
	  penalty = penalty + WDcount;
	  flapCount = flapCount + flapWDcount;
	  flapNum = flapNum + flapWDcount;
	}
      }
      WDcount = 0;
      flapWDcount = 0;
      prevDop = curDop;
      twoBits = curTwoBits;
    }
  }
  else {   //Traditional RFD implementation
    //flap:10
    if(withdraw && prevUpdate == 1) {
      penalty += withdrawPenalty;
      flapCount++;
      flapNum++;
    }
    //flap:11(implicit_withdrawal)
    else if(!withdraw && prevUpdate == 1) {
      penalty += changePenalty;
      flapCount++;
      flapNum++;
    }
    //flap:01
    else if(!withdraw && prevUpdate == 0) {
      //Cisco doesn't punish readvertisement after withdrawal
      if(bgp->rfd_punish_readvertisement) {  //if readvertisement after withdrawal is punished
	penalty += withdrawPenalty; //the penalty is the same as a withdrawal
	flapCount++;
        flapNum++;
      }
    }
  }
  if(penalty > penalty_ceiling) {
    penalty = penalty_ceiling;
  }
  //make sure we don't suppress until fourth flap
  if(penalty > suppressLimit && flapCount >= 4) {
    bool tmp = suppressed;
    suppressed = true;
    if(!tmp) {
      supNum++;
      printf("\ntotal number of suppresses so far is: %d\n", supNum);
    }
  }
  else if(penalty <= reuseLimit + 0.1) {
    suppressed = false;
  }
  //cancel reuse timer if set
  if(reuseTimer != NULL && reuseTimer->scheduled()) {
    reuseTimer->cancel();
  }
  //if suppressed and route reachable, set up reuse timer
  if(suppressed) {
    double coeff;
    if(reachable) {
      coeff = log(2.0) / halfLife_ok;
    } else {
      coeff = log(2.0) / halfLife_ng;
    }
    double supTime = (log(penalty) - log(reuseLimit)) / coeff;
    //make sure we don't suppress longer than maximum suppress time
    if(supTime > maxSuppressTime) {
      supTime = maxSuppressTime;
    }
    //set up the reuse timer
    reuseTimer = new DampReuseTimer(bgp, supTime, this);
    bgp->set_timer(reuseTimer);
    //printf("Reuse Timer Set: %f\n", supTime);
  }
  if(withdraw == true) {  //remember current update
    prevUpdate = 0 ;  
  } else {
    prevUpdate = 1;  
  }
  routeInfo = ri;  //remember routeinfo
  printf("\ntotal number of flaps so far is: %d\n", flapNum); 
}

/** Display some related information */
string DampInfo::toString() {
  string str = "";
  char* temp = new char[20];
  str = str + nlri->toString() + ":\n";
  str = str + "    recorded penalty: ";
  sprintf(temp, "%f", penalty);
  str = str + string(temp) + ";";
  str = str + "  number of flaps so far: ";
  sprintf(temp, "%d", flapCount); 
  str = str + string(temp) + ";";
  str = str + "  suppressed: ";
  if(suppressed)  str = str + "yes;";
  else  str = str + "no;";
  str = str + "  peer: " + peerNh.substr(peerNh.find_first_of(":")+1) + "\n";
  return str;
}

